define(['jquery', 'underscore', 'backbone', 'tpls', 'context'],
    function ($, _, Backbone, Templates, Context) {
      var SortingView = Backbone.View.extend({
        className: "row filter-sorting-row",
        template: Templates["content/query/filter/sorting"],
        events: {
          'click .remove': 'removeView',
          'change :input': 'change'
        },
        initialize: function () {
          _.bindAll(this, "focus");
          this.listenTo(this.model, 'change', this.render);
          this.listenTo(this.model, 'remove', this.removeView);
        },
        removeView: function () {
          this.model.destroy();
          this.remove();
        },
        change: function (e) {
          var $input = $(e.currentTarget);
          // silence model update because the event in coming from dom and the elements are
          // already up to date, otherwise this will cause double render
          this.model.set($input.attr('name'), $input.val(), {silent: true});
        },
        focus: function () {
          this.$("input[type=text],textarea,select").first().focus();
        },
        /**
         * Set input values back to model
         */
        sync: function (model) {
          this.$(":input").each(function () {
            var $input = $(this);
            if ($input.attr("name")) {
              model.set($input.attr("name"), $input.val())
            }
          });
        },
        render: function () {
          var data = {
            model: this.model.toJSON(),
            columns: Context.meta.columns,
            orders: ["ASC", "DESC"]
          };
          var tpl = this.template(data);
          this.$el.html(tpl);

          this.sync(this.model);

          return this;
        }
      });
      return SortingView;
    });