var $el = $("#chart");
var data = {};
var settings = {};
var plot;

function isNumber(n) {
    return !isNaN(parseFloat(n)) && isFinite(n);
}

function showTooltip(x, y, contents) {
    $("<div class='tooltip'><div class='tooltip-inner'>" + contents + "</div></div>").css({
        position: "absolute",
        display: "none",
        top: y + 7,
        left: x + 7,
        opacity: 0.90
    }).appendTo("body").fadeIn(200);
}

$(function () {

    $.extend(settings, local_config);

    var ticks = settings.labelField ? [] : null;

    $.get(url, function (res) {
        $.each(res['rows'], function (i, item) {
            $.each(item, function (j, cell) {
                if (isNumber(cell)) {
                    if (!data.hasOwnProperty(j)) {
                        data[j] = [];
                    }
                    data[j].push([i, parseFloat(cell)]);

                    if (settings.labelField && res['rows'][i][settings.labelField] !== undefined) {
                        ticks.push([i, res['rows'][i][settings.labelField]]);
                    }
                }
            });
        });

        // map data for charting
        var map_data = $.map(data, function (v, k) {
            return {label: k, data: v};
        });

        console.log("map_data:", map_data);

        // show warning if there is not enough data to draw a chart
        if (map_data.length == 0) {
            alert("Not enough data to draw a chart!");
            return;
        }

        // plot chart
        plot = $.plot($el, map_data, {
            series: {
                bars: {
                    show: true,
                    barWidth: 0.8,
                    align: "center"
                },
                points: {
                    show: true
                }
            },
            grid: {
                autoHighlight: true,
                hoverable: true,
                clickable: false,
                tickColor: "#e8e8e8",
                borderWidth: 1,
                borderColor: "#ccc"
            },
            xaxis: {
                tickDecimals: 0,
                ticks: ticks,
                labelWidth: 30
            },
            yaxis: {
                tickDecimals: 0,
                autoscaleMargin: 0.1
            }
        });

        // get y axis position
        var yaxis_left = $(".flot-y-axis>div:first").position().left;

        // add title
        $el.append($("<div>")
            .addClass("chart-title badge")
            .css("left", yaxis_left + 20) // relative to y axis
            .text(res.title));

        var previousPoint = null;
        $el.bind("plothover", function (event, pos, item) {
            if (item) {
                if (previousPoint != item.dataIndex) {

                    previousPoint = item.dataIndex;

                    $(".tooltip").remove();
                    var x = item.datapoint[0].toFixed(2),
                        y = item.datapoint[1].toFixed(2);

                    var text = "Value: ";
                    text += y;

                    showTooltip(item.pageX, item.pageY, text);
                }
            } else {
                $(".tooltip").remove();
                previousPoint = null;
            }
        });
    }, "json");
});